
/*----------------------------------------------------------------------+
 |									|
 |	robot.c								|
 |									|
 +----------------------------------------------------------------------*/

/*
 *  robot client for icfp 2002 programming contest
 *  http://icfpcontest.cse.ogi.edu/
 *  31-Aug-2002 Marcel van Kervinck
 */

/*----------------------------------------------------------------------+
 |	standard includes						|
 +----------------------------------------------------------------------*/

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*----------------------------------------------------------------------+
 |	system includes							|
 +----------------------------------------------------------------------*/

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>

#include <unistd.h>
#include <sys/uio.h>

#include <sys/time.h>
#include <sys/resource.h>

#define RUSAGE_SELF             0
#define RUSAGE_CHILDREN         -1

/*----------------------------------------------------------------------+
 |									|
 |	utils 								|
 |									|
 +----------------------------------------------------------------------*/

static char *program;

typedef int bool;
#define FALSE 0
#define TRUE 1

struct xy {
	short x;
	short y;
};

#define BUF 16384

static
char *format(const char *fmt, ...)
{
	static char line[BUF];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(line, sizeof line, fmt, ap);
	va_end(ap);

	return line;
}

#define ASSERT(cond,mesg)\
do{\
	if(!(cond)) {\
		fprintf(stderr,\
			"%s: error: %s\n%s: error: file %s, line %d\n",\
			program, format mesg,\
			program, __FILE__, __LINE__);\
		exit(EXIT_FAILURE);\
	}\
}while(0)

#define INF 1000000

/*----------------------------------------------------------------------+
 |	function: xclock						|
 +----------------------------------------------------------------------*/

static
double xclock(void)
{
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru)) {
		return -1.0;
	}

	return (double)(ru.ru_utime.tv_usec + ru.ru_stime.tv_usec)/1000000
		+ (ru.ru_utime.tv_sec + ru.ru_stime.tv_sec);
}

/*----------------------------------------------------------------------+
 |      simple random generator                                         |
 +----------------------------------------------------------------------*/

static int rnd_seed = 1;

static int rnd(void)
{
	int r = rnd_seed;

	r = 16807 * (r % 127773) - 2836 * (r / 127773);
	if (r < 0) r += 0x7fffffff;

	return rnd_seed = r;
}

/*----------------------------------------------------------------------+
 |									|
 |	communication							|
 |									|
 +----------------------------------------------------------------------*/

static char *com_buffer;
static int com_buflen = 0;
static int com_bufmax = 0;

static int com_fd;

/*----------------------------------------------------------------------+
 |	function: com_start						|
 +----------------------------------------------------------------------*/

static
void com_start(char *address, int port)
{
	int b0, b1, b2, b3;
	struct sockaddr_in sa;
	struct hostent *hp;
	int r;

#ifdef DEBUG
	printf("address=%s, port=%d\n", address, port);
#endif

	if (!(hp = gethostbyname(address))) {
		r = sscanf(address, "%d.%d.%d.%d", &b0, &b1, &b2, &b3);
		ASSERT(r==4, ("failed to resolve address"));

		hp = calloc(1, sizeof(*hp));
		hp->h_addrtype = AF_INET;
		hp->h_length = 4;
		hp->h_addr_list = (char **) calloc (2, sizeof(char *));
		hp->h_addr_list[0] = (char *) malloc (4);
		hp->h_addr_list[0][0] = b0;
		hp->h_addr_list[0][1] = b1;
		hp->h_addr_list[0][2] = b2;
		hp->h_addr_list[0][3] = b3;
	}
	bzero(&sa, sizeof(struct sockaddr_in));
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons((unsigned short) port);
	memcpy(&sa.sin_addr, hp->h_addr, hp->h_length);

	com_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	ASSERT(com_fd>=0, ("socket() failed"));

	r = connect(com_fd, (struct sockaddr *) &sa, sizeof(struct sockaddr_in));
	ASSERT(r>=0, ("connect() failed"));
}

/*----------------------------------------------------------------------+
 |	function: com_readline						|
 +----------------------------------------------------------------------*/

static
void com_readline(void)
{
	char ch;
	int i;

	static char buf[BUF];
	static int pos = 0;

	com_buflen = 0;

	for (;;) {
		for (i=0; i<pos; i++) {
			ch = buf[i];

			ASSERT(ch != '\0', ("unexpected byte in input (%d)", (int)ch));

			/* make sure there is space to add something to com_buffer */

			if (com_buflen >= com_bufmax) {
				com_bufmax += com_bufmax/2;
				if (com_bufmax < 4096) com_bufmax = 4096;

				com_buffer = realloc(com_buffer, com_bufmax);
				ASSERT(com_buffer!=NULL, ("realloc() for com_buffer failed"));
			}

			if (ch == '\n') {
				/* done! re-adjust buf[] and return */

				pos -= i+1;
				bcopy(buf+i+1, buf, pos);

				com_buffer[com_buflen] = '\0';
				goto sorry_about_the_goto;
			}


			com_buffer[com_buflen] = ch;
			com_buflen+=1;
		}

		/* by definition, buf[] is exhausted now, so read more */

		pos = read(com_fd, buf, BUF);
		if (pos == 0) {
			/* no more data */
			exit(0);
		}
		ASSERT(pos >= 0, ("read() return with return value %d", pos));
	}

sorry_about_the_goto:
	printf("com_readline: '%s'\n", com_buffer);
}

/*----------------------------------------------------------------------+
 |	function: com_write						|
 +----------------------------------------------------------------------*/

static
void com_write(const char *text)
{
	int r;

	r = write(com_fd, text, strlen(text));
	ASSERT(r==strlen(text), ("write() failed"));

	printf("com_write: '%s'\n", text);
}

/*----------------------------------------------------------------------+
 |									|
 |	game								|
 |									|
 +----------------------------------------------------------------------*/

#define MAP_BASE		'@'
#define MAP_EMPTY		'.'
#define MAP_WATER		'~'
#define MAP_WALL		'#'

#define MAP_MAXX		1024
#define MAP_MAXY		1024

static int map_maxx;
static int map_maxy;

static int game_capacity;
static int game_money;

static int game_move_nr;

static int my_robot_id;
static struct robot *my_robot;

static int my_money;
static int my_weight;
static int my_score;

/*
 *  the current plan
 */
static int my_target_x;		/* where to go */
static int my_target_y;
static int my_target_pick;	/* package id: only set if the plan is to go and pick a package */

static char map[MAP_MAXY][MAP_MAXX];

/*----------------------------------------------------------------------+
 |	dead-lock / repetition detection				|
 +----------------------------------------------------------------------*/

#define REP_HISTORY 12
#define REP_MAX_REPEAT 4

static struct xy rep_history[REP_HISTORY];
static int rep_index;

static
int cmp_xy(const void *pa, const void *pb)
{
	struct xy const *a = pa;
	struct xy const *b = pb;

	if (a->y < b->y) return -1;
	if (a->y > b->y) return 1;

	if (a->x < b->x) return -1;
	if (a->x > b->x) return 1;

	return 0;
}

static
bool rep_test(void)
{
	struct xy h[REP_HISTORY];
	int i;
	int best_count = 0;
	int count;
	struct xy prev = { 0, 0 };

	memcpy(h, rep_history, sizeof(rep_history));
	for (i=0; i<REP_HISTORY; i++) {
		if (h[i].x == 0) {
			h[i].x = -i-1;	/* makes empty slots unique */
		}
	}

	qsort(h, REP_HISTORY, sizeof(h[0]), cmp_xy);

	count = 0;
	for (i=0; i<REP_HISTORY; i++) {
		if (cmp_xy(&prev, &h[i]) == 0) {	/* the same */
			count++;
		} else {			/* different */
			prev = h[i];
			if (count > best_count) {
				best_count = count;
			}
			count = 1;
		}
	}

	if (count > best_count) {
		best_count = count;
	}

	return (best_count >= REP_MAX_REPEAT);
}

static
void rep_record_pos(int x, int y)
{
	rep_history[rep_index].x = x;
	rep_history[rep_index].y = y;

	rep_index = (rep_index + 1) % REP_HISTORY;
}

static
void rep_reset(void)
{
	int i;

	for (i=0; i<REP_HISTORY; i++) {
		rep_history[i].x = 0;
		rep_history[i].y = 0;
	}
}

/*----------------------------------------------------------------------+
 |	packages							|
 +----------------------------------------------------------------------*/

struct package {
	int			id;
	int			dest_x;
	int			dest_y;
	int			weight;

	int			x;
	int			y;

	/* packages are linked in a robot list, or in world */
	struct package		*next;
	struct package		**prev;

	bool			mark;
};

static struct package *package_world;	/* packages that are not in a robot */
static int nr_packages;

static
struct package *new_package(int id, int x, int y)
{
	struct package *pac;

	pac = malloc(sizeof(*pac));
	ASSERT(pac!=NULL, ("no memory for new package"));

	pac->id = id;
	pac->dest_x = x;
	pac->dest_y = y;
	pac->weight = 1;

	pac->x = x;
	pac->y = y;

	/* a new package is always inserted into world first */

	pac->next = package_world;
	if (package_world != NULL) {
		package_world->prev = &pac->next;
	}
	pac->prev = &package_world;

	pac->mark = FALSE;

	package_world = pac;

	nr_packages += 1;

	return pac;
}

/*----------------------------------------------------------------------+
 |	robots								|
 +----------------------------------------------------------------------*/

struct robot {
	int			id;
	int			x;
	int			y;

	struct package		*pacs;

	bool			mark;
};

static struct robot **robots;
static int nr_robots;

/*----------------------------------------------------------------------+
 |	function: find_robot						|
 +----------------------------------------------------------------------*/

static
struct robot *find_robot(int id)
{
	struct robot *bot = NULL;
	int i;

	for (i=0; i<nr_robots; i++) {
		if (robots[i]->id == id) {
			bot = robots[i];
			break;
		}
	}

	return bot;
}

/*----------------------------------------------------------------------+
 |	function: new_robot						|
 +----------------------------------------------------------------------*/

static
struct robot *new_robot(int id, int x, int y)
{
	struct robot *bot;

	ASSERT(find_robot(id)==NULL, ("duplicate robot id=%d", id));

	bot = malloc(sizeof(*bot));
	ASSERT(bot!=NULL, ("no memory for new robot(id=%d,...)", id));

	bot->id = id;
	bot->x = x;
	bot->y = y;

	bot->pacs = NULL;

	bot->mark = FALSE;

	/* add it to the list */

	nr_robots += 1;
	robots = realloc(robots, sizeof(robots[0]) * nr_robots);
	ASSERT(robots!=NULL, ("realloc() failed while adding new robot id=%d to robots-list", id));
	robots[nr_robots-1] = bot;

	if (id == my_robot_id) {
		my_robot = bot;
	}

	return bot;
}

/*----------------------------------------------------------------------+
 |	function: robot_move						|
 +----------------------------------------------------------------------*/

#define DIR_NORTH		'N'
#define DIR_SOUTH		'S'
#define DIR_EAST		'E'
#define DIR_WEST		'W'

static
int robot_move(struct robot *bot, char dir)
{
	int newx, newy;
	int result = 0;

	newx = bot->x;
	newy = bot->y;

	switch (dir) {
	case DIR_NORTH:
		newy += 1;
		break;
	case DIR_SOUTH:
		newy -= 1;
		break;
	case DIR_EAST:
		newx += 1;
		break;
	case DIR_WEST:
		newx -= 1;
		break;
	default:
		ASSERT(FALSE, ("bad direction '%c'", dir));
	}

	switch (map[newy][newx]) {
	case MAP_EMPTY:
		/* ok */
		break;
	case MAP_WALL:
		newx = bot->x;
		newy = bot->y;
		result = 1;
		break;
	case MAP_WATER:
		/* should die */
		newx = bot->x;
		newy = bot->y;
		result = 2;
		break;
	default:
		ASSERT(FALSE, ("map[y=%d][x=%d] corrupt, contains %d", newy, newx, (int)map[newy][newx]));
	}

	bot->x = newx;
	bot->y = newy;

	return result;
}

/*----------------------------------------------------------------------+
 |	function: compute_distances					|
 +----------------------------------------------------------------------*/

static int dist[MAP_MAXY][MAP_MAXX];
static int dist_curr_x;
static int dist_curr_y;

static
void compute_distances(int to_x, int to_y)
{
	int x, y;
	int d;

	struct xy *list = NULL;
	int ix1 , ix2, ix3;

	if ((to_x == dist_curr_x) && (to_y == dist_curr_y)) {
		return;	/* already done */
	}
	dist_curr_x = to_x;
	dist_curr_y = to_y;

	/* init, water and walls are given +infinity  */

	for (y=0; y<=map_maxy+1; y++) {
		for (x=0; x<=map_maxx+1; x++) {
			if (map[y][x] == MAP_EMPTY) {
				dist[y][x] = -1; /* unknown */
			} else {
				dist[y][x] = INF; /* no entry */
			}
		}
	}

	list = malloc(sizeof(list[0]) * map_maxx * map_maxy);
	ASSERT(list!=NULL, ("no memory for list"));

	d  = 0;
	ix3 = 0;
	dist[to_y][to_x] = d;
	list[ix3].x = to_x;
	list[ix3].y = to_y;
	ix3++;

	ix1 = 0;
	do {
		d += 1;
		ix2 = ix3;

		/* find back all locations with dist = d-1 */
		while (ix1 < ix2) {
			int xx, yy;

			/* loop over their neighbors and mark new neighbors as dist d */

			xx = list[ix1].x - 1;
			yy = list[ix1].y;
			if (dist[yy][xx] < 0) {
				dist[yy][xx] = d;
				list[ix3].x = xx;
				list[ix3].y = yy;
				ix3++;
			}

			xx = list[ix1].x + 1;
			yy = list[ix1].y;
			if (dist[yy][xx] < 0) {
				dist[yy][xx] = d;
				list[ix3].x = xx;
				list[ix3].y = yy;
				ix3++;
			}

			xx = list[ix1].x;
			yy = list[ix1].y - 1;
			if (dist[yy][xx] < 0) {
				dist[yy][xx] = d;
				list[ix3].x = xx;
				list[ix3].y = yy;
				ix3++;
			}

			xx = list[ix1].x;
			yy = list[ix1].y + 1;
			if (dist[yy][xx] < 0) {
				dist[yy][xx] = d;
				list[ix3].x = xx;
				list[ix3].y = yy;
				ix3++;
			}

			ix1++;
		}

	} while (ix2 < ix3);

#if 0
	for (y=0; y<=map_maxy+1; y++) {
		for (x=0; x<=map_maxx+1; x++) {
			int ch = dist[y][x];

			if (ch == INF) ch = ' ';
			else ch = 33 + ch%63;
			putchar(ch);
		}
		printf("\n");
	}
	sleep(3);

#endif

	free(list);
}

/*----------------------------------------------------------------------+
 |	function: p2p_distance						|
 +----------------------------------------------------------------------*/

#define DIST_CACHE (1<<20)
#define DIST_EXTRA 7

struct cache {
	struct xy p1;
	struct xy p2;
	int d;
	int used;
};

static struct cache dist_cache[DIST_CACHE + DIST_EXTRA];

static int cache_misses;
static int cache_hits;

static
int p2p_distance(int x1, int y1, int x2, int y2)
{
	int d;
	unsigned int hash;
	struct cache *p, *q;
	int i;

	/* normalize */

	if ((y1 > y2) || ( (y1==y2) && (x1 > x2) )) {
		int tmp;
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	/* lookup */

	hash = 0xdeadbeef;
	hash = hash * 23 + y1;
	hash = hash * 23 + x1;
	hash = hash * 23 + y2;
	hash = hash * 23 + x2;
	hash &= DIST_CACHE-1;

	p = &dist_cache[hash];

	q = NULL;
	for (i=0; i<=DIST_EXTRA; i++) {
		if (	(p[i].p1.x == x1) &&
			(p[i].p1.y == y1) &&
			(p[i].p2.x == x2) &&
			(p[i].p2.y == y2))
		{
			/* found */

			cache_hits++;

			p[i].used++;
			d = p[i].d;

			if (p[i].used > INF) {
#ifdef DEBUG
	printf("*** SCALING DISTANCE CACHE ***\n");
#endif
				for (i=0; i<DIST_CACHE+DIST_EXTRA; i++) {
					dist_cache[i].used >>= 2;
				}
			}

			return d;
		}
		if ((q == NULL) || ( (p->used < q->used) )) {
			q = p;
		}
	}

	cache_misses++;

	/* compute */

	compute_distances(x1, y1);

	d = dist[y2][x2];

	if (d < 0) {
		d = +INF;
	}

	/* store */

	ASSERT(q!=NULL, ("cannot store d"));

	q->p1.x = x1;
	q->p1.y = y1;
	q->p2.x = x2;
	q->p2.y = y2;
	q->d = d;
	q->used = 0;

	return d;
}

/*----------------------------------------------------------------------+
 |	function: game_display						|
 +----------------------------------------------------------------------*/

static
int cmp_robot(const void *pa, const void *pb)
{
	struct robot * const *a = pa;
	struct robot * const *b = pb;

	if ((*a)->y < (*b)->y) return -1;
	if ((*a)->y > (*b)->y) return 1;
	if ((*a)->x < (*b)->x) return -1;
	if ((*a)->x > (*b)->x) return 1;
	if ((*a)->id < (*b)->id) return -1;
	if ((*a)->id > (*b)->id) return 1;
	return 0;
}

static
void game_display()
{
#ifdef DEBUG
	int x, y, ch;
#endif
	struct robot **bots, **bot;

	bots = malloc(sizeof(bots[0]) * nr_robots);
	ASSERT(bots!=NULL, ("malloc failed"));

	memcpy(bots, robots, sizeof(bots[0]) * nr_robots);
	qsort(bots, nr_robots, sizeof(bots[0]), cmp_robot);
	bot = bots;

#if 0
	for (ch=0; ch<nr_robots; ch++) {
		if (bots[ch] != NULL) {
			printf("id=%d y=%d x=%d\n",
				bots[ch]->id,
				bots[ch]->y,
				bots[ch]->x);
		} else {
			printf("NULL\n");
		}
	}
#endif
#if 0
	printf("\e[1;1H");
#endif
#ifdef DEBUG
	for (y=0; y<=map_maxy+1; y++) {
		for (x=0; x<=map_maxx+1; x++) {
			ch = map[y][x];

			if (bot<&bots[nr_robots])
			if (((*bot)->y == y) && ((*bot)->x == x)) {
				if ((*bot)->id == my_robot_id) {
					ch = 'X';
				} else {
					ch = 'O';
				}
				bot++;
			}
			fputc(ch, stdout);
		}
		printf("\n");
	}
#endif
#if 0
	ASSERT(bot == &bots[nr_robots], ("robot missing"));
#endif

	printf("x=%d, y=%d, money=%d, score=%d, weight=%d, capacity=%d, robots=%d, packages=%d\n",
		my_robot->x,
		my_robot->y,
		my_money,
		my_score,
		my_weight,
		game_capacity,
		nr_robots,
		nr_packages);

	printf("target x=%d y=%d (%d)\n", my_target_x, my_target_y, my_target_pick);

#ifdef DEBUG
	{struct package *p;
		for (p=my_robot->pacs; p!=NULL; p=p->next) {
			printf("holding id=%3d dest (%3d, %3d), w=%d\n",
				p->id,
				p->dest_x, p->dest_y,
				p->weight);
		}
	}
#endif


	free(bots);
}

/*----------------------------------------------------------------------+
 |	function: game_start						|
 +----------------------------------------------------------------------*/

static
void game_start(void)
{
	int x, y;
	int r;
	int i;

	rnd_seed = time(NULL);

	com_write("Player\n");
	com_readline();
	/* get max_x and max_Y */

	r = sscanf(com_buffer, "%d%d",
		&map_maxx,
		&map_maxy);
	ASSERT(r==2, ("expected map dimensions x, y, but read '%s'", com_buffer));

	/*
	 *  read map
	 */
	for (y=1; y<=map_maxy; y++) {
		com_readline();

		for (x=1; x<=map_maxx; x++) {
			char ch = com_buffer[x-1];

			switch (ch) {
			case MAP_WALL:
			case MAP_EMPTY:
			case MAP_WATER:
				break;
			case MAP_BASE:
				/* create a dummy package (id=-1) in world */
				(void) new_package(-1, x, y);

				ch = MAP_EMPTY;
				break;
			default:
				ASSERT(FALSE, ("reading invalid map element '%c'", ch));
				break;
			}
			map[y][x] = ch;
		}
	}

	/*
	 *  draw surrounding walls
	 *
	 *  1111111111
	 *  3........4
	 *  3........4
	 *  3........4
	 *  2222222222
	 */
	for (x=0; x<=map_maxx+1; x++) {
		map[0][x] = MAP_WALL;			/* 1 */
		map[map_maxy+1][x] = MAP_WALL;		/* 2 */
	}
	for (y=1; y<=map_maxy; y++) {
		map[y][0] = MAP_WALL;			/* 3 */
		map[y][map_maxx+1] = MAP_WALL;		/* 4 */
	}

	/*
	 *  read robot id, capacity, money
	 */
	com_readline();
	r = sscanf(com_buffer, "%d%d%d",
		&my_robot_id,
		&game_capacity,
		&game_money);
	ASSERT(r==3, ("expected 3 numbers, but read '%s'", com_buffer));

	my_money = game_money;

	/*
	 *  read robot locations
	 */
	com_readline();

	i = 0;
	while (com_buffer[i] != '\0') {
		int n;
		int x, y, id;
		r = sscanf(com_buffer+i, "#%d X %d Y %d %n", &id, &x, &y, &n);
		ASSERT(r==3, ("sscanf() failed for reading robot locations, input: '%s'", com_buffer+i));
#ifdef DEBUG
		printf("robot id=%d x=%d y=%d\n", id, x, y);
#endif
		(void) new_robot(id, x, y);
		i += n;
	}
}

/*----------------------------------------------------------------------+
 |	function: game_select_target					|
 +----------------------------------------------------------------------*/

/*
 *  monte-carlo simulation to select a new target
 *  (either walking to get a package, or delivering one that we are holding)
 */
#define PICK_NONE (-INF)

static
void game_select_target(void)
{
	int best_score = 0;

	int iterations = 10000;

	struct package **pac_holding = NULL;
	struct package **pac_world = NULL;

	/*
	 *  init with just a random valid location, in case we find nothing
	 *  better to do later
	 */
	do {
		my_target_x = 1 + (rnd() % map_maxx);
		my_target_y = 1 + (rnd() % map_maxy);
		my_target_pick = PICK_NONE;
	} while ( (map[my_target_y][my_target_x] != MAP_EMPTY) ||
		( p2p_distance(my_target_x, my_target_y, my_robot->x, my_robot->y) >= +INF ));

	/*
	 *  alloc memory for package lists
	 */
	pac_holding = malloc(sizeof(pac_holding[0]) * nr_packages);
	ASSERT(pac_holding!=NULL, ("no memory"));

	pac_world = malloc(sizeof(pac_holding[0]) * nr_packages);
	ASSERT(pac_world!=NULL, ("no memory"));

	/*
	 *  monte-carlo
	 */

	do {
		int first_target_x = 0;
		int first_target_y = 0;
		int first_target_pick = PICK_NONE;

		int score = 0;
		int money = my_money;

		int x = my_robot->x;
		int y = my_robot->y;
		int pick;

		int newx, newy;

		int w = 0;

		int nr_pac_holding;

		int nr_pac_world;

		struct package *p;

		/*
		 *  create lists of packages in robot and in world
		 */

		nr_pac_holding = 0;
		for (p = my_robot->pacs; p!=NULL; p=p->next) {
			w += p->weight;
			pac_holding[nr_pac_holding] = p;
			nr_pac_holding++;
		}

		ASSERT(w == my_weight, ("weight mismatch, w=%d, my_weight=%d", w, my_weight));

		nr_pac_world = 0;
		for (p = package_world; p!=NULL; p=p->next) {
			pac_world[nr_pac_world] = p;
			nr_pac_world++;
		}

		/*
		 *  simulate until the end
		 */
		while (money > 0) {
			int best_rand = -INF;
			int curr_rand;

			int action = '?';
			int pac_ix = -1;

			int i;

			/*
		 	 *  randomly select an action
		 	 */

			pick = PICK_NONE;

			for (i=0; i<nr_pac_holding; i++) {
				/*
				 *  option 1: deliver this package
				 */

				curr_rand = rnd();
				if (curr_rand > best_rand) {
					best_rand = curr_rand;

					action = 'D';
					pac_ix = i;
					pick = PICK_NONE;
				}
			}

			for (i=0; i<nr_pac_world; i++) {
				/*
				 *  option 2: get this one, only if we have the capacity
				 */

				if (w + pac_world[i]->weight <= game_capacity) {
					curr_rand = rnd();
					if (curr_rand > best_rand) {
						best_rand = curr_rand;

						action = 'F';
						pac_ix = i;

						pick = pac_world[i]->id;
						ASSERT(pick!=PICK_NONE, ("uhm"));
					}

				}
			}


			/*
			 *  if there is no move, stop this run
			 */
			if (best_rand < 0) {
				break;
			}

			/*
			 *  execute the selected action
			 */

			switch (action) {
			case 'D':
				/*
				 *  move robot to destination
				 */
				newx = pac_holding[pac_ix]->dest_x;
				newy = pac_holding[pac_ix]->dest_y;
				money -= p2p_distance(x, y, newx, newy);
				money -= 1;
				x = newx;
				y = newy;

				/*
				 *  deliver package
				 */
				score += 100000 * pac_holding[pac_ix]->weight / (1 + my_money - money);
				w -= pac_holding[pac_ix]->weight;

				/*
				 *  remove package from robot list
				 */
				nr_pac_holding--;
				pac_holding[pac_ix] = pac_holding[nr_pac_holding];


				break;

			case 'F':
				/*
				 *  move robot to destination
				 */
				newx = pac_world[pac_ix]->x;
				newy = pac_world[pac_ix]->y;
				money -= p2p_distance(x, y, newx, newy);
				money -= 1;
				x = newx;
				y = newy;

				/*
				 *  add package to robot
				 */
				pac_holding[nr_pac_holding] = pac_world[pac_ix];
				w += pac_holding[nr_pac_holding]->weight;
				ASSERT(w <= game_capacity, ("internal error, robot overload"));
				nr_pac_holding++;

				/*
				 *  remove package from world
				 */
				nr_pac_world--;
				pac_world[pac_ix] = pac_world[nr_pac_world];

				break;

			default:
				ASSERT(FALSE, ("internal error"));
				break;
			}

			/*
			 *  if this is the first move, remember what it was
			 */
			if (first_target_x == 0) {
				first_target_x = x;
				first_target_y = y;
				first_target_pick = pick;
			}
		}

		/*
		 *  if this scenario is better, keep it
		 */

		if (score > best_score) {
			best_score = score;

			my_target_x = first_target_x;
			my_target_y = first_target_y;
			my_target_pick = first_target_pick;
		}

		if ((iterations & 127) == 0) {
			double cpu = xclock();
			/*
			 *  set target at at most of 0.5 sec cpu/move,
			 *  but asking the first 5.0 seconds anyhow
			 */
			if (cpu-5.0 > game_move_nr/2.0) {
#ifdef DEBUG
				printf("*** RUNNING OUT OF TIME (cpu=%1.3f, move_nr=%d)\b\n", cpu, game_move_nr);
#endif
				break;
			}
		}

	} while (iterations-- > 0);

#ifdef DEBUG
	printf("***\n");
	printf("*** strategy: target x=%d y=%d (pick=%d)\n", my_target_x, my_target_y, my_target_pick);
	printf("***\n");
#endif
	free(pac_holding);
	free(pac_world);
}

/*----------------------------------------------------------------------+
 |	function: game_process_package_line				|
 +----------------------------------------------------------------------*/

static
int game_process_package_line(void)
{
	struct package *p;
	int i;
	int changed = 0;

	/*
	 *  first mark packages in world that we think should
	 *  be here
	 */

	for (p = package_world; p!=NULL; p=p->next) {
		if ((p->x == my_robot->x) && (p->y == my_robot->y)) {
			p->mark = TRUE;
		}
	}

	/*
	 *  what items are on current location?
	 */
	com_readline();

	if (1 == sscanf(com_buffer, "Robot %d", &i)) {	/* 'Robot id died' -> game over */
		if (i == my_robot_id) {
			exit(0);
		}
		return 0; /* what ever! */
	}


	i = 0;
	while (com_buffer[i] != '\0') {
		int n, r;
		int id, x, y, w;

		r = sscanf(com_buffer+i, "%d%d%d%d %n", &id, &x, &y, &w, &n);
		ASSERT(r==4, ("expecting to read 4 integers, but got '%s'", com_buffer+i));
		i += n;

		ASSERT(id!=PICK_NONE, ("package id in world == %d", id));

		/*
		 *  find this package in the world
		 */

		for (p=package_world; p!=NULL; p=p->next) {
			if (p->id == id) {
				/* ok, do not worry */
				ASSERT(p->mark, ("mark expected to be set, but it isnt"));
				p->mark = FALSE;
				break;
			}
		}
		if (p==NULL) {
			/*
			 *  package not yet known found: create one
			 */

			p = new_package(id, x, y);
			p->x = my_robot->x;
			p->y = my_robot->y;
			p->weight = w;

			changed++;
		}
	}

	/*
	 *  now look again at the packages if we miss one
	 */

	p = package_world;
	while (p!=NULL) {
		struct package *q = p->next;

		if (p->mark) {
			/*
			 *  this one is not present in the real world: delete it
			 *  (could be a dummy)
			 */
			if (p->next != NULL) {
				p->next->prev = p->prev;
			}
			*(p->prev) = p->next;
#ifdef DEBUG
			printf ("deleting package id=%d\n", p->id);
#endif

			free(p);
			nr_packages--;

			changed++;
		}
		p = q;
	}

	return changed;
}

/*----------------------------------------------------------------------+
 |	function: game_search_escape					|
 +----------------------------------------------------------------------*/

#define FIXBID(bid) ((abs(bid) < my_money) ? bid : ((bid < 0) ? -1 : 1))

/*
 *  does a small tree-search to find the best move.
 *  this method is only called when there are opponent robots very nearby
 *  and we are next to water.
 */

static
void game_search_escape(char default_move, int dist)
{
	int oldx = my_robot->x;
	int oldy = my_robot->y;

	int best_bid = 50;
	int best_move = default_move;
	int best_score = 1;
	int i, j;
	int r;

	char buf[BUF];

	/*
	 *  attempt to move away from any robot
	 */
	for (i=0; i<4; i++) {
		char action = i["NSEW"];

		r = robot_move(my_robot, action);
		if (r == 0) {
			/* ok, valid move */

			bool near_water = (
				(map[my_robot->y-1][my_robot->x] == MAP_WATER) ||
				(map[my_robot->y+1][my_robot->x] == MAP_WATER) ||
				(map[my_robot->y][my_robot->x-1] == MAP_WATER) ||
				(map[my_robot->y][my_robot->x+1] == MAP_WATER) );

			/* find nearest robot */
			int score = INF;
			for (j=0; j<nr_robots; j++) {
				if (robots[j] != my_robot) {
					int d = p2p_distance(
						my_robot->x,
						my_robot->y,
						robots[j]->x,
						robots[j]->y);
					if (d < score) {
						score = d;
					}
				}
			}

			if (near_water && (score == 1)) {
				score = 0;
			}
			if (score >= best_score) {
				if (dist > 2) {
					best_bid = 1;
				} else
				if (dist == 2) {
					if (score <= 1) {
						best_bid = 10;
					} else {
						best_bid = 1;
					}
				} else {
					best_bid = 50;
				}
				best_move = action;
				best_score = score;
			}
		} else {
			/* not a good move */
		}

		/* restore */

		my_robot->x = oldx;
		my_robot->y = oldy;
	}

#ifdef DEBUG
	printf("*** ESCAPING: bid=%d best_score=%d\n", best_bid, best_score);
#endif

	best_bid = FIXBID(best_bid);
	sprintf(buf, "%d Move %c\n", best_bid, best_move);
	com_write(buf);
	my_money -= abs(best_bid);

	return;
}

/*----------------------------------------------------------------------+
 |	function: game_make_move					|
 +----------------------------------------------------------------------*/

/*
 *  make one move closer to the target, or select a new target
 *  (or drop or pick package)
 */

static
void game_make_move(void)
{
	struct robot *bot = my_robot;
	int r;

	int score, best;
	int curr_rand, best_rand;
	char action = '?';
	int bid = 1;
	char buf[BUF];

	int i;
	int ddd;

	int oldx, oldy;
	struct package *p;
	int buf_ix;

	/*
	 *  inspect packages at current location
	 *  if anything has changed, then reset the current plan so we
	 *  can make a new one
	 */

	r = game_process_package_line();
	if (r > 0) {
		/* world changed, reset plan */
		my_target_x = 0;
		my_target_y = 0;
		my_target_pick = 0;
	}

	/*
	 *  if we are at the destination of one of the packages that we are
	 *  holding: drop the most valuable one. (irrespective of the actual
	 *  plan, which we must reset)
	 */

	buf_ix = 0;
	for (p=my_robot->pacs; p!=NULL; p=p->next) {
		if ((my_robot->x == p->dest_x) && (my_robot->y == p->dest_y)) {
			if (buf_ix == 0) {
				bid = FIXBID(bid);
				buf_ix += sprintf(buf+buf_ix, "%d Drop", bid);
			}

			buf_ix += sprintf(buf+buf_ix, " %d", p->id);
		}
	}
	if (buf_ix > 0) {
		sprintf(buf+buf_ix, "\n");
		/*
		 *  @@@ this is potentially dangerous in the presence of
		 *  other robots...
		 */
		com_write(buf);
		my_money -= abs(bid);

		my_target_x = 0;
		my_target_y = 0;
		my_target_pick = 0;

		return;
	}

	/*
	 *  deadlock/repetition: whatever the plan: if we have a repeat, we change plans and to
	 *  walk somewhere else
	 */

	if (rep_test()) {
		do {
			my_target_x = 1 + (rnd() % map_maxx);
			my_target_y = 1 + (rnd() % map_maxy);
			my_target_pick = PICK_NONE;
		} while ( (map[my_target_y][my_target_x] != MAP_EMPTY) ||
			( p2p_distance(my_target_x, my_target_y, my_robot->x, my_robot->y) > 20 ));

		rep_reset();
#ifdef DEBUG
	printf("*** HELP! MOVING SOMEWHERE ELSE\n");
#endif
	}

	/*
	 *  if the plan is empty, we must make a new plan
	 */

	if (my_target_x == 0) {
		game_select_target();
	}

	/*
	 *  if on target to pick a package, do so and reset the plan
	 */
	ASSERT(my_target_x > 0, ("internal error: no plan"));

	while ((my_robot->x == my_target_x) &&
		(my_robot->y == my_target_y))
	{
		if (my_target_pick == PICK_NONE) {
			/*
			 *   ok, out of options. there are no packages that
			 *   we are aware of. started walking randomly now,
			 *   try another place.
			 */
#ifdef DEBUG
			printf("RANDOM RUNNER\n");
#endif
			game_select_target();
			continue;

		}
#if 0
		ASSERT(my_target_pick != PICK_NONE, ("robot on target location forgot what to pick"));
#endif

		/*
		 *  @@@ this also is potentially dangerous
		 */

		bid = FIXBID(bid);
		sprintf(buf, "%d Pick %d\n", bid, my_target_pick);
		com_write(buf);
		my_money -= abs(bid);

		my_target_x = 0;
		my_target_y = 0;
		my_target_pick = 0;

		return;
	}

	/*
	 *  we are executing a plan: travelling towards a target
	 *  move one step
	 */

	compute_distances(my_target_x, my_target_y);

	best = +INF;
	best_rand = 0;
	oldx = bot->x;
	oldy = bot->y;

	score = dist[oldy+1][oldx];
	if (score <= best) {
		curr_rand = best_rand;
		if ((score < best) || ((curr_rand=rnd()) > best_rand )) {
			action = 'N';
			best = score;
			best_rand = curr_rand;
		}
	}

	score = dist[oldy-1][oldx];
	if (score <= best) {
		curr_rand = best_rand;
		if ((score < best) || ((curr_rand=rnd()) > best_rand )) {
			action = 'S';
			best = score;
			best_rand = curr_rand;
		}
	}

	score = dist[oldy][oldx-1];
	if (score <= best) {
		curr_rand = best_rand;
		if ((score < best) || ((curr_rand=rnd()) > best_rand )) {
			action = 'W';
			best = score;
			best_rand = curr_rand;
		}
	}

	score = dist[oldy][oldx+1];
	if (score <= best) {
		curr_rand = best_rand;
		if ((score < best) || ((curr_rand=rnd()) > best_rand )) {
			action = 'E';
			best = score;
			best_rand = curr_rand;
		}
	}

	/*
	 *  now check if there are any enemy robots very nearby.
	 *  if so, we do a local search deal with them
	 */

	for (i=0; i<nr_robots; i++) {
		if ((robots[i]->id != my_robot_id) &&
			((ddd=p2p_distance(
				my_robot->x,
				my_robot->y,
				robots[i]->x,
				robots[i]->y) <= 3)))
		{
			/* got one! alarm */
#ifdef DEBUG
			printf ("*** ALERT: ROBOT %d NEARBY ***\n",
				robots[i]->id);
#endif
			/*
			 *  alarm near water means searching an escape
			 */
			if ((map[my_robot->y-1][my_robot->x] == MAP_WATER) ||
			    (map[my_robot->y+1][my_robot->x] == MAP_WATER) ||
			    (map[my_robot->y][my_robot->x-1] == MAP_WATER) ||
			    (map[my_robot->y][my_robot->x+1] == MAP_WATER) ||
			    (map[my_robot->y-1][my_robot->x-1] == MAP_WATER) ||
			    (map[my_robot->y-1][my_robot->x+1] == MAP_WATER) ||
			    (map[my_robot->y+1][my_robot->x-1] == MAP_WATER) ||
			    (map[my_robot->y+1][my_robot->x+1] == MAP_WATER))
			{
				game_search_escape(action,ddd);
				return;
			}
		}
	}

	/*
	 *  all is safe: move
	 */

	bid = FIXBID(bid);
	sprintf(buf, "%d Move %c\n", bid, action);
	com_write(buf);
	my_money -= abs(bid);
}

/*----------------------------------------------------------------------+
 |	function: game_process_moves					|
 +----------------------------------------------------------------------*/

static
void game_process_moves(void)
{
	int i;

	bool made_pick = FALSE;

	com_readline();

	for (i=0; i<nr_robots; i++) {
		robots[i]->mark = TRUE;
	}

	i = 0;
	while (com_buffer[i] != '\0') {
		int id;
		int ch;
		int r, n;
		int x, y;
		int pac_id;
		struct robot *bot = NULL;

		r = sscanf(com_buffer+i, "#%d %n", &id, &n);
		ASSERT(r==1, ("sscanf() failed for robot id. got '%s'", com_buffer+i));

		bot = find_robot(id);

		if (bot != NULL) {
			bot->mark = FALSE;
		}

		i += n;
		ch = com_buffer[i];

		while ((ch != '#') && (ch != '\0')) {
			switch (ch) {
			case 'X':
				r = sscanf(com_buffer+i, "X %d Y %d %n", &x, &y, &n);
				ASSERT(r==2, ("expected X and Y, but got '%s'", com_buffer+i));
				i += n;
				bot = find_robot(id);
				if (bot != NULL) {
					bot->x = x;
					bot->y = y;
				} else {
					bot = new_robot(id, x, y);
				}
				break;

			case DIR_NORTH:
			case DIR_SOUTH:
			case DIR_WEST:
			case DIR_EAST:
				ASSERT(bot!=NULL, ("unknown robot id=%d", id));
				r = robot_move(bot, ch);

#ifdef DEBUG
	printf("robot id=%d (moving %c) is now at x=%d y=%d (result=%d)\n", bot->id, ch, bot->x, bot->y, r);
#endif


				r = sscanf(com_buffer+i, "%*s %n", &n);
				ASSERT(r==0, ("error skipping direction, got '%s'"));
				i += n;
				break;

			case 'P':
				r = sscanf(com_buffer+i, "P %d %n", &pac_id, &n);
				ASSERT(r==1, ("expected P and number, but got '%s'", com_buffer+i));
				i += n;

				if (id == my_robot_id) {
					struct package *p, *q;
					/*
					 *  transfer package from world into robot list
					 */

					made_pick = TRUE;

					p=package_world;
					while (p!=NULL) {
						q = p->next;
						if (p->id == pac_id) {
							if (p->next != NULL) {
								p->next->prev = p->prev;
							}
							*(p->prev) = p->next;

							if (my_robot->pacs != NULL) {
								my_robot->pacs->prev = &(p->next);
							}
							p->next = my_robot->pacs;
							p->prev = &my_robot->pacs;
							my_robot->pacs = p;

#ifdef DEBUG
	printf("my robot picked: package id %d, dest x=%d y=%d\n", p->id, p->dest_x, p->dest_y);
#endif

							my_weight += p->weight;
						}
						p = q;
					}

				} else {
					struct package *p, *q;
					/*
					 *  another robot made a pick: if we know the package,
					 *  forget about its existence
					 *  also reset any plan
					 */
					p=package_world;
 					while (p!=NULL) {
						q = p->next;
						if (p->id == pac_id) {
							if (p->next != NULL) {
								p->next->prev = p->prev;
							}
							*(p->prev) = p->next;
							free(p);
							nr_packages--;

							my_target_x = 0;
							my_target_y = 0;
							my_target_pick = PICK_NONE;
						}
						p = q;
					}

				}
				break;

			case 'D':
				r = sscanf(com_buffer+i, "D %d %n", &pac_id, &n);
				ASSERT(r==1, ("expected D and number, but got '%s'", com_buffer+i));
				i += n;

				if (id == my_robot_id) {
					struct package *p, *q;

					/*
					 *  remove package it from the robot list
					 */
					p=my_robot->pacs;
					while (p!=NULL) {
						q = p->next;
						if (p->id == pac_id) {
							/*
							 *  we may have scored
							 */
							if ((p->dest_x == my_robot->x) &&
							    (p->dest_y == my_robot->y)) {
								my_score += p->weight;
							}

							if (p->next != NULL) {
								p->next->prev = p->prev;
							}
							*(p->prev) = p->next;
#if DEBUG
	printf("removed package id=%d from robot list\n", p->id);
#endif
							my_target_x = 0;
							my_target_y = 0;
							my_target_pick = PICK_NONE;

							my_weight -= p->weight;
							free(p);
							nr_packages--;
						}
						p = q;
					}
				}

				break;

			default:
				ASSERT(FALSE, ("go away ('%s')", com_buffer+i));
				break;
			}

			/* next char */
			ch = com_buffer[i];
		}
	}

	if (!made_pick) {
		rep_record_pos(my_robot->x, my_robot->y);
	}

	i = 0;
	while (i < nr_robots) {
		if (robots[i]->mark) {
#ifdef DEBUG
			printf("*** removing robot id=%d\n", robots[i]->id);
#endif
			free(robots[i]);
			nr_robots--;
			robots[i] = robots[nr_robots];
			robots = realloc(robots, sizeof(robots[0]) * nr_robots);
			ASSERT(robots!=NULL, ("realloc() failed while removing dead robot"));
		} else {
			i++;
		}
	}
}

/*----------------------------------------------------------------------+
 |									|
 |	main								|
 |									|
 +----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
	char *address = "icfp1.cse.ogi.edu";
	int port = 22009;

	int argi = 1;

	/*
	 *  process arguments
	 */

	program = argv[0];

	if (argi < argc) {
		address = argv[argi++];
	}

	if (argi < argc) {
		port = atoi(argv[argi++]);
	}

	if (argi < argc) {
		fprintf(stderr, "%s: too many arguments\n", program);
		exit(EXIT_FAILURE);
	}

	/*
	 *  connect to server
	 */
	com_start(address, port);

	/*
	 *  process game configuration
	 */
	game_start();

	/*
 	 *  play the game
	 */
	for (;;) {
		game_move_nr += 1;

#if 0
		{ char xxx[BUF]; gets(xxx); }
#endif
#ifdef DEBUG
		printf("\n*** move nr %d\n", game_move_nr);
#endif

		/*
		 *  select move
		 */
		game_make_move();

		game_display();

		/*
		 *  process move results
		 */
		game_process_moves();

	}
	return 0;
}

/*----------------------------------------------------------------------+
 |									|
 +----------------------------------------------------------------------*/

